.. _vp{{ versiondata.version.replace(".", "p") }}:

{{ "v" + versiondata.version }}
{{ top_underline * ((versiondata.version)|length + 1) }}
{% for section, _ in sections.items() %}
{% set underline = underlines[0] %}

{% if sections[section] %}
{% for category, category_data in definitions.items() if category in sections[section] %}
{{ category_data['name'] }}
{{ underline * category_data['name']|length }}
{% if category_data['showcontent'] %}
{% for text, issues in sections[section][category].items() %}
{% set lines = text.strip().splitlines() %}
- {{ lines[0] }} {% if issues %}({{ issues|join(', ') }}){% endif %}

{% if lines|length > 1 %}
  {{ lines[1:]|join('\n  ') }}
{% endif %}
{% endfor %}

{% else %}
- {{ sections[section][category]['']|join(', ') }}

{% endif %}
{% endfor %}
{% else %}
No significant changes.


{% endif %}
{% endfor %}
